#include "RS485.hpp"

// C++ system level
#include <cinttypes>

// ESP32 specific
#include "esp_log.h"
#include "driver/gpio.h"
#include "driver/uart.h"

// project specific
// #include "LambdaTask.hpp"

// qthing stuff
// #include ""



static const char* TAG = "RS485";


// This leads to ~54 P/s @ 2 MBaud at 100% load
// TODO: Make configurable?!
const uint32_t CyanBusMTU = 4096;


namespace SiliconTorch {

  namespace Impl {


    RS485::RS485(uint8_t tx, uint8_t rx, uint8_t de, uint8_t re, uint32_t baudRate, uint8_t uartChannel) : tx(tx), rx(rx), de(de), re(re), uartChannel(uartChannel) {

      uint64_t bitMask = 0;

      bitMask |= 1 << tx;
      bitMask |= 1 << de;
      bitMask |= 1 << re;

      gpio_config_t gpio = {
        .pin_bit_mask = bitMask,
        .mode         = GPIO_MODE_OUTPUT,
        .pull_up_en   = GPIO_PULLUP_DISABLE,
        .pull_down_en = GPIO_PULLDOWN_DISABLE,
        .intr_type    = GPIO_INTR_DISABLE,        
      };

      // configure outputs
      gpio_config(&gpio);
      txrxEN(false);

      gpio.pin_bit_mask = 1 << rx;
      gpio.mode = GPIO_MODE_INPUT;

      // configure inputs
      gpio_config(&gpio);


      uart_config_t uartConfig = {
        .baud_rate           = (int)baudRate,
        .data_bits           = UART_DATA_8_BITS,
        .parity              = UART_PARITY_DISABLE,
        .stop_bits           = UART_STOP_BITS_1,
        .flow_ctrl           = UART_HW_FLOWCTRL_DISABLE,
        .rx_flow_ctrl_thresh = 0,
        .use_ref_tick        = false,
        // .source_clk = UART_SCLK_APB,
      };


      uart_port_t _ch = (uart_port_t)uartChannel;

      uart_driver_install(_ch, CyanBusMTU, 0, 0, NULL, 0);
      uart_param_config(_ch, &uartConfig);
      uart_set_pin(_ch, tx, rx, UART_PIN_NO_CHANGE, UART_PIN_NO_CHANGE);
      uart_set_mode(_ch, UART_MODE_UART);
      uart_set_rx_timeout(_ch, 0);  // TODO: timeout…??

      // TOOD: receive only ATM!
      txEN(false);
      rxEN(true);
    }


    void RS485::write(const uint8_t* data, uint32_t length) {
      uint32_t send = uart_write_bytes((uart_port_t)uartChannel, reinterpret_cast<const char*>(data), length);

      // TODO: remove after change!
      ESP_LOGE(TAG, "This class is receive-only atm! TX won't work!");

      if (send != length) ESP_LOGE(TAG, "Sending bytes#[ %d ] failed: Only bytes#[ %d ] were send!", length, send);
    }

    void RS485::write(std::string& data) {
      uint32_t send = uart_write_bytes((uart_port_t)uartChannel, data.c_str(), data.length());

      // TODO: remove after change!
      ESP_LOGE(TAG, "This class is receive-only atm! TX won't work!");

      if (send != data.length()) ESP_LOGE(TAG, "Sending bytes#[ %d ] failed: Only bytes#[ %d ] were send!", data.length(), send);
    }

    // returns the real amount of data read
    uint32_t RS485::read(uint8_t* data, uint32_t length) {
      return uart_read_bytes((uart_port_t)uartChannel, data, length, 0);  // TODO: ticks…???
    }

    uint32_t RS485::available() {

      size_t tmp;
      uart_get_buffered_data_len((uart_port_t)uartChannel, &tmp);

      return (uint32_t)tmp;
    }

    void RS485::setBaudRate(uint32_t baudRate) {
      uart_set_baudrate((uart_port_t)uartChannel, baudRate);
    }

    uint32_t RS485::getBaudRate() {
      uint32_t baudRate = 0;
      uart_get_baudrate((uart_port_t)uartChannel, &baudRate);
      return baudRate;
    }


    void RS485::txEN(bool state) {
      gpio_set_level((gpio_num_t)de, state);
    }

    void RS485::rxEN(bool state) {
      gpio_set_level((gpio_num_t)re, state ^ 1);
    }
    
    void RS485::txrxEN(bool state) {
      txEN(state);
      rxEN(state);
    }

  }
}



